//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop
#include "App_DiskImages.h"
#include "Moa_AutoPointers.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
//---------------------------------------------------------------------------
namespace App_diskimages {
//---------------------------------------------------------------------------
int __fastcall TDiskImageFormat::ReadInteger(void* AData, int AOffset, int ASize)
{
	if(ASize <= 0 || ASize > 4) {
		throw EDiskImageError("The program cannot read the integer.");
	}
	int LValue = 0;
	for(int LI = 0; LI < ASize; ++LI) {
		LValue |= static_cast<int>(*(static_cast<byte*>(AData) + AOffset + LI)) << (8 * LI);
	}
	return LValue;
}
//---------------------------------------------------------------------------
void __fastcall TDiskImageFormat::WriteInteger(void* AData, int AOffset, int ASize, int AValue)
{
	if(ASize <= 0 || ASize > 4) {
		throw EDiskImageError("The program cannot write the integer.");
	}
	for(int LI = 0; LI < ASize; ++LI) {
		*(static_cast<byte*>(AData) + AOffset + LI) = (AValue >> (8 * LI)) & 0xFF;
	}
}
//---------------------------------------------------------------------------
AnsiString __fastcall TDiskImageFormat::ReadString(void* AData, int AOffset, int ASize)
{
	if(ASize < 1) {
		throw EDiskImageError("The program cannot read the string.");
	}
	TAutoArray<char> LBuffer = new char[ASize];
	::ZeroMemory(LBuffer, ASize);
	if(::lstrcpyn(LBuffer, static_cast<byte*>(AData) + AOffset, ASize) == NULL) {
		throw EDiskImageError("The program could not read the string.");
	}
	return AnsiString(LBuffer);
}
//---------------------------------------------------------------------------
void __fastcall TDiskImageFormat::WriteString(void* AData, int AOffset, int ASize, const AnsiString& AValue)
{
	if(ASize < 1 || AValue.Length() > ASize - 1) {
		throw EDiskImageError("The program cannot write the string.");
	}
	::ZeroMemory(static_cast<byte*>(AData) + AOffset, ASize);
	if(::lstrcpyn(static_cast<byte*>(AData) + AOffset, AValue.c_str(), ASize) == NULL) {
		throw EDiskImageError("The program could not write the string.");
	}
}
//---------------------------------------------------------------------------
const AnsiString& __fastcall TDiskImageFormat::AssureValidName(const AnsiString& AName)
{
	if(!IsValidName(AName)) {
		throw EDiskImageError(AnsiString().sprintf(
			"The program cannot accept the name \"%s\".",
			AName.c_str()));
	}
	return AName;
}
//---------------------------------------------------------------------------
bool __fastcall TDiskImageFormat::IsValidName(const AnsiString& AName)
{
	return AName.Length() <= TDiskImageFormat::MaxNameLength;
}
//---------------------------------------------------------------------------
int __fastcall TDiskImageFormat::ToggleWriteProtected(int AWriteProtected)
{
	switch(AWriteProtected) {
	case WriteProtected_Yes:
		return WriteProtected_No;
	case WriteProtected_No:
		return WriteProtected_Yes;
	default:
		throw EAppError("The program cannot accept the wirte-protected.");
	}
}
//---------------------------------------------------------------------------
int __fastcall TDiskImageFormat::AssureValidWriteProtected(int AWriteProtected)
{
	if(!IsValidWriteProtected(AWriteProtected)) {
		throw EDiskImageError("The program cannot accept the write-protected.");
	}
	return AWriteProtected;
}
//---------------------------------------------------------------------------
bool __fastcall TDiskImageFormat::IsValidWriteProtected(int AWriteProtected)
{
	switch(AWriteProtected) {
	case WriteProtected_Yes:
	case WriteProtected_No:
		return true;
	default:
		return false;
	}
}
//---------------------------------------------------------------------------
int __fastcall TDiskImageFormat::AssureValidDiskType(int ADiskType)
{
	if(!IsValidDiskType(ADiskType)) {
		throw EDiskImageError("The program cannot accept the disk type.");
	}
	return ADiskType;
}
//---------------------------------------------------------------------------
bool __fastcall TDiskImageFormat::IsValidDiskType(int ADiskType)
{
	switch(ADiskType) {
	case DiskType_2D:
	case DiskType_2DD:
	case DiskType_2HD:
		return true;
	default:
		return false;
	}
}
//---------------------------------------------------------------------------
int __fastcall TDiskImageFormat::AssureValidSize(int ASize)
{
	if(!IsValidSize(ASize)) {
		throw EDiskImageError("The program cannot accept the size.");
	}
	return ASize;
}
//---------------------------------------------------------------------------
bool __fastcall TDiskImageFormat::IsValidSize(int ASize)
{
	return ASize >= HeaderSize;
}
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
void __fastcall TDiskImage::Initialize(int ADiskType)
{
	try {
		FData = new TMemoryStream();
		Create(ADiskType);
		Update();
	}
	catch(...) {
		Uninitialize();
		throw;
	}
}
//---------------------------------------------------------------------------
void __fastcall TDiskImage::Uninitialize()
{
	delete FData;
	FData = NULL;
}
//---------------------------------------------------------------------------
void __fastcall TDiskImage::Create(int ADiskType)
{
	TAutoPointer<TMemoryStream> LBuffer = new TMemoryStream();
	LBuffer->Size = TDiskImageFormat::HeaderSize;
	::ZeroMemory(LBuffer->Memory, LBuffer->Size);
	TDiskImageFormat::WriteInteger(
		LBuffer->Memory,
		TDiskImageFormat::WriteProtectedOffset,
		TDiskImageFormat::WriteProtectedSize,
		TDiskImageFormat::WriteProtected_No);
	TDiskImageFormat::WriteInteger(
		LBuffer->Memory,
		TDiskImageFormat::DiskTypeOffset,
		TDiskImageFormat::DiskTypeSize,
		TDiskImageFormat::AssureValidDiskType(ADiskType));
	TDiskImageFormat::WriteInteger(
		LBuffer->Memory,
		TDiskImageFormat::SizeOffset,
		TDiskImageFormat::SizeSize,
		LBuffer->Size);
	Data->WriteBuffer(LBuffer->Memory, LBuffer->Size);
}
//---------------------------------------------------------------------------
void __fastcall TDiskImage::Update()
{
	Name = TDiskImageFormat::ReadString(
		Data->Memory,
		TDiskImageFormat::NameOffset,
		TDiskImageFormat::NameSize);
	DiskType = TDiskImageFormat::ReadInteger(
		Data->Memory,
		TDiskImageFormat::DiskTypeOffset,
		TDiskImageFormat::DiskTypeSize);
	WriteProtected = TDiskImageFormat::ReadInteger(
		Data->Memory,
		TDiskImageFormat::WriteProtectedOffset,
		TDiskImageFormat::WriteProtectedSize);
	SetSize(TDiskImageFormat::ReadInteger(
		Data->Memory,
		TDiskImageFormat::SizeOffset,
		TDiskImageFormat::SizeSize));
}
//---------------------------------------------------------------------------
void __fastcall TDiskImage::UpdateData()
{
	TDiskImageFormat::WriteString(
		Data->Memory,
		TDiskImageFormat::NameOffset,
		TDiskImageFormat::NameSize,
		Name);
	TDiskImageFormat::WriteInteger(
		Data->Memory,
		TDiskImageFormat::DiskTypeOffset,
		TDiskImageFormat::DiskTypeSize,
		DiskType);
	TDiskImageFormat::WriteInteger(
		Data->Memory,
		TDiskImageFormat::WriteProtectedOffset,
		TDiskImageFormat::WriteProtectedSize,
		WriteProtected);
	TDiskImageFormat::WriteInteger(
		Data->Memory,
		TDiskImageFormat::SizeOffset,
		TDiskImageFormat::SizeSize,
		Size);
}
//---------------------------------------------------------------------------
void __fastcall TDiskImage::SetName(const AnsiString& AName)
{
	FName = TDiskImageFormat::AssureValidName(AName);
}
//---------------------------------------------------------------------------
void __fastcall TDiskImage::SetDiskType(int ADiskType)
{
	FDiskType = TDiskImageFormat::AssureValidDiskType(ADiskType);
}
//---------------------------------------------------------------------------
void __fastcall TDiskImage::SetWriteProtected(int AWriteProtected)
{
	FWriteProtected = TDiskImageFormat::AssureValidWriteProtected(AWriteProtected);
}
//---------------------------------------------------------------------------
void __fastcall TDiskImage::SetSize(int ASize)
{
	FSize = TDiskImageFormat::AssureValidSize(ASize);
}
//---------------------------------------------------------------------------
__fastcall TDiskImage::TDiskImage(int ADiskType) :
	TObject(),
	FData(NULL),
	FName(),
	FDiskType(),
	FWriteProtected(),
	FSize()
{
	Initialize(ADiskType);
}
//---------------------------------------------------------------------------
__fastcall TDiskImage::~TDiskImage()
{
	Uninitialize();
}
//---------------------------------------------------------------------------
void __fastcall TDiskImage::Assign(TDiskImage* ASource)
{
	if(ASource == this) {
		return;
	}
	FName = ASource->FName;
	FDiskType = ASource->FDiskType;
	FWriteProtected = ASource->FWriteProtected;
	FSize = ASource->FSize;
	FData->Size = 0;
	FData->WriteBuffer(ASource->FData->Memory, ASource->FData->Size);
}
//---------------------------------------------------------------------------
void __fastcall TDiskImage::LoadFromFile(const AnsiString& AFile)
{
	try {
		TAutoPointer<TMemoryStream> LStream = new TMemoryStream();
		LStream->LoadFromFile(AFile);
		LStream->Position = 0;
		LoadFromStream(LStream);
	}
	catch(EFOpenError& LException) {
		LException.Message = AnsiString().sprintf(
			"The program could not open the file: \"%s\".",
			AFile.c_str());
		throw;
	}
	catch(EInvalidDiskImage& LException) {
		LException.Message = AnsiString().sprintf(
			"The program could not load a disk image from the file: \"%s\".",
			AFile.c_str());
		throw;
	}
}
//---------------------------------------------------------------------------
void __fastcall TDiskImage::SaveToFile(const AnsiString& AFile)
{
	try {
		TAutoPointer<TMemoryStream> LStream = new TMemoryStream();
		SaveToStream(LStream);
		LStream->SaveToFile(AFile);
	}
	catch(EFOpenError& LException) {
		LException.Message = AnsiString().sprintf(
			"The program could not open the file: \"%s\".",
			AFile.c_str());
		throw;
	}
	catch(EInvalidDiskImage& LException) {
		LException.Message = AnsiString().sprintf(
			"The program could not save the disk image to the file: \"%s\".",
			AFile.c_str());
		throw;
	}
}
//---------------------------------------------------------------------------
void __fastcall TDiskImage::LoadFromStream(TStream* AStream)
{
	try {
		TAutoPointer<TDiskImage> LInstance = new TDiskImage();
		{
			// Read the header
			LInstance->Data->Size = TDiskImageFormat::HeaderSize;
			LInstance->Data->Position = LInstance->Data->Size;
			AStream->ReadBuffer(LInstance->Data->Memory, LInstance->Data->Size);

			// Read the tracks
			{
				int LSize = TDiskImageFormat::AssureValidSize(TDiskImageFormat::ReadInteger(
					LInstance->Data->Memory,
					TDiskImageFormat::SizeOffset,
					TDiskImageFormat::SizeSize));
				int LRest = LSize - TDiskImageFormat::HeaderSize;
				TAutoPointer<TMemoryStream> LBuffer = new TMemoryStream();
				LBuffer->Size = 1024 * 32;
				while(LRest > 0) {
					int LReadSize = LBuffer->Size;
					if(LRest < LReadSize) {
						LReadSize = LRest;
					}
					AStream->ReadBuffer(LBuffer->Memory, LReadSize);
					LInstance->Data->WriteBuffer(LBuffer->Memory, LReadSize);
					LRest -= LReadSize;
				}
			}
			LInstance->Update();
		}
		Assign(LInstance);
	}
	catch(EReadError& LException) {
		throw EInvalidDiskImage("The program could not load the disk image.");
	}
	catch(EDiskImageError& LException) {
		throw EInvalidDiskImage("The program could not load the disk image.");
	}
}
//---------------------------------------------------------------------------
void __fastcall TDiskImage::SaveToStream(TStream* AStream)
{
	try {
		UpdateData();
		AStream->Write(Data->Memory, Data->Size);
	}
	catch(EWriteError& LException) {
		throw EInvalidDiskImage("The program could not save the disk image.");
	}
	catch(EDiskImageError& LException) {
		throw EInvalidDiskImage("The program could not save the disk image.");
	}
}
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
int __fastcall TDiskImageDescriber::EvaluateWriteProtected(int AWriteProtected)
{
	switch(AWriteProtected) {
	case TDiskImageFormat::WriteProtected_No:
		return 1;
	case TDiskImageFormat::WriteProtected_Yes:
		return 0;
	default:
		throw EDiskImageError("The program cannot accept the write-protected.");
	}
}
//---------------------------------------------------------------------------
int __fastcall TDiskImageDescriber::EvaluateDiskType(int ADiskType)
{
	switch(ADiskType) {
	case TDiskImageFormat::DiskType_2D:
		return 0;
	case TDiskImageFormat::DiskType_2DD:
		return 1;
	case TDiskImageFormat::DiskType_2HD:
		return 2;
	default:
		throw EDiskImageError("The program cannot accept the disk type.");
	}
}
//---------------------------------------------------------------------------
int TDiskImageDescriber::ImageIndex_WriteProtected_Yes = -1;
int TDiskImageDescriber::ImageIndex_WriteProtected_No = -1;
//---------------------------------------------------------------------------
AnsiString __fastcall TDiskImageDescriber::DescribeByName(TDiskImage* ADiskImage)
{
	return ADiskImage->Name;
}
//---------------------------------------------------------------------------
AnsiString __fastcall TDiskImageDescriber::DescribeByWriteProtected(TDiskImage* ADiskImage)
{
	switch(ADiskImage->WriteProtected) {
	case TDiskImageFormat::WriteProtected_No:
		return "No";
	case TDiskImageFormat::WriteProtected_Yes:
		return "Yes";
	default:
		throw EDiskImageError("The program cannot accept the write-protected.");
	}
}
//---------------------------------------------------------------------------
AnsiString __fastcall TDiskImageDescriber::DescribeByDiskType(TDiskImage* ADiskImage)
{
	switch(ADiskImage->DiskType) {
	case TDiskImageFormat::DiskType_2D:
		return "2D";
	case TDiskImageFormat::DiskType_2DD:
		return "2DD";
	case TDiskImageFormat::DiskType_2HD:
		return "2HD";
	default:
		throw EDiskImageError("The program cannot accept the disk type.");
	}
}
//---------------------------------------------------------------------------
AnsiString __fastcall TDiskImageDescriber::DescribeBySize(TDiskImage* ADiskImage)
{
	return TCommon::FormatInteger(ADiskImage->Size);
}
//---------------------------------------------------------------------------
int __fastcall TDiskImageDescriber::CompareByName(TDiskImage* ADiskImage1, TDiskImage* ADiskImage2)
{
	return AnsiCompareStr(
		ADiskImage1->Name,
		ADiskImage2->Name);
}
//---------------------------------------------------------------------------
int __fastcall TDiskImageDescriber::CompareByWriteProtected(TDiskImage* ADiskImage1, TDiskImage* ADiskImage2)
{
	return
		EvaluateWriteProtected(ADiskImage1->WriteProtected) -
		EvaluateWriteProtected(ADiskImage2->WriteProtected);
}
//---------------------------------------------------------------------------
int __fastcall TDiskImageDescriber::CompareByDiskType(TDiskImage* ADiskImage1, TDiskImage* ADiskImage2)
{
	return
		EvaluateDiskType(ADiskImage1->DiskType) -
		EvaluateDiskType(ADiskImage2->DiskType);
}
//---------------------------------------------------------------------------
int __fastcall TDiskImageDescriber::CompareBySize(TDiskImage* ADiskImage1, TDiskImage* ADiskImage2)
{
	return
		ADiskImage1->Size -
		ADiskImage2->Size;
}
//---------------------------------------------------------------------------
int __fastcall TDiskImageDescriber::DetermineImageIndex(TDiskImage* ADiskImage)
{
	if(ADiskImage->WriteProtected) {
		return ImageIndex_WriteProtected_Yes;
	}
	else {
		return ImageIndex_WriteProtected_No;
	}
}
//---------------------------------------------------------------------------
} // End namespace App_diskimages
//---------------------------------------------------------------------------

