//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop
#include "Moa_CommandPattern.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
//---------------------------------------------------------------------------
// Version: 2.0.0
// Date: 2012-07-12 UTC+09:00
// Author: Moochos
//---------------------------------------------------------------------------
namespace Moa_commandpattern {
//---------------------------------------------------------------------------
__fastcall TCommand::TCommand() :
	TObject(),
	FResults()
{
	// Do nothing
}
//---------------------------------------------------------------------------
__fastcall TCommand::~TCommand()
{
	// Do nothing
}
//---------------------------------------------------------------------------
void __fastcall TCommand::Undo()
{
	throw ECommandError("The command cannot accept the undo operation.");
}
//---------------------------------------------------------------------------
void __fastcall TCommand::Redo()
{
	throw ECommandError("The command cannot accept the redo operation.");
}
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
void __fastcall TInvoker::Initialize()
{
	try {
		FQueue = new TObjectList();
		FHistory = new TObjectList();
	}
	catch(...) {
		Uninitialize();
		throw;
	}
}
//---------------------------------------------------------------------------
void __fastcall TInvoker::Uninitialize()
{
	delete FHistory;
	FHistory = NULL;

	delete FQueue;
	FQueue = NULL;
}
//---------------------------------------------------------------------------
void __fastcall TInvoker::FinishExecute()
{
	if(Next->Results.Contains(crClearQueue)) {
		Queue->Count = 1;
	}
	if(Next->Results.Contains(crClearUndo)) {
		UndoCount = 0;
	}
	if(Next->Results.Contains(crClearRedo)) {
		RedoCount = 0;
	}
	if(!Next->Results.Contains(crCanUndo) || HistoryLimit == 0) {
		Queue->Delete(0);
	}
	else {
		RedoCount = 0;
		if(History->Count == HistoryLimit) {
			UndoCount = UndoCount / 2;
		}
		History->Add(Next);
		Queue->Extract(Next);
		++FPosition;
	}
}
//---------------------------------------------------------------------------
void __fastcall TInvoker::FinishUndo()
{
	if(NextUndo->Results.Contains(crClearQueue)) {
		Queue->Count = 0;
	}
	if(NextUndo->Results.Contains(crClearUndo)) {
		UndoCount = 1;
	}
	if(NextUndo->Results.Contains(crClearRedo)) {
		RedoCount = 0;
	}
	if(NextUndo->Results.Contains(crCanRedo)) {
		--FPosition;
	}
	else {
		--FPosition;
		RedoCount = 0;
	}
}
//---------------------------------------------------------------------------
void __fastcall TInvoker::FinishRedo()
{
	if(NextRedo->Results.Contains(crClearQueue)) {
		Queue->Count = 0;
	}
	if(NextRedo->Results.Contains(crClearUndo)) {
		UndoCount = 0;
	}
	if(NextRedo->Results.Contains(crClearRedo)) {
		RedoCount = 1;
	}
	if(NextRedo->Results.Contains(crCanUndo)) {
		++FPosition;
	}
	else {
		++FPosition;
		UndoCount = 0;
	}
}
//---------------------------------------------------------------------------
TCommand* __fastcall TInvoker::GetNext()
{
	return static_cast<TCommand*>(Queue->First());
}
//---------------------------------------------------------------------------
int __fastcall TInvoker::GetCount()
{
	return Queue->Count;
}
//---------------------------------------------------------------------------
TCommand* __fastcall TInvoker::GetNextUndo()
{
	return static_cast<TCommand*>(History->Items[Position - 1]);
}
//---------------------------------------------------------------------------
int __fastcall TInvoker::GetUndoCount()
{
	return Position;
}
//---------------------------------------------------------------------------
void __fastcall TInvoker::SetUndoCount(int AUndoCount)
{
	if(AUndoCount < 0 || AUndoCount > UndoCount) {
		throw ECommandError("The program cannot set the number of the undo commands.");
	}
	if(AUndoCount == UndoCount) {
		return;
	}
	int LDifference = UndoCount - AUndoCount;
	for(int LI = 0; LI < History->Count - LDifference; ++LI) {
		History->Exchange(LI, LDifference + LI);
	}
	History->Count = History->Count - LDifference;
	FPosition -= LDifference;
	if(Origin >= 0) {
		if(Origin < LDifference) {
			FOrigin = -1;
		}
		else {
			FOrigin -= LDifference;
		}
	}
}
//---------------------------------------------------------------------------
TCommand* __fastcall TInvoker::GetNextRedo()
{
	return static_cast<TCommand*>(History->Items[Position]);
}
//---------------------------------------------------------------------------
int __fastcall TInvoker::GetRedoCount()
{
	return History->Count - Position;
}
//---------------------------------------------------------------------------
void __fastcall TInvoker::SetRedoCount(int ARedoCount)
{
	if(ARedoCount < 0 || ARedoCount > RedoCount) {
		throw ECommandError("The program cannot set the number of the redo commands.");
	}
	if(ARedoCount == RedoCount) {
		return;
	}
	History->Count = History->Count - (RedoCount - ARedoCount);
	if(Origin > History->Count) {
		FOrigin = -1;
	}
}
//---------------------------------------------------------------------------
int __fastcall TInvoker::GetHistorySize()
{
	return UndoCount + RedoCount;
}
//---------------------------------------------------------------------------
void __fastcall TInvoker::SetHistoryLimit(int AHistoryLimit)
{
	int LHistoryLimit = AHistoryLimit;
	if(LHistoryLimit < 0) {
		LHistoryLimit = -1;
	}
	if(LHistoryLimit == HistoryLimit) {
		return;
	}
	if(LHistoryLimit >= 0) {
		int LDifference = History->Count - LHistoryLimit;
		if(RedoCount > LDifference) {
			RedoCount = RedoCount - LDifference;
		}
		else {
			LDifference -= RedoCount;
			RedoCount = 0;
			UndoCount = UndoCount - LDifference;
		}
	}
	FHistoryLimit = LHistoryLimit;
}
//---------------------------------------------------------------------------
void __fastcall TInvoker::DoExecuting()
{
	if(OnExecuting != NULL) {
		OnExecuting(this);
	}
}
//---------------------------------------------------------------------------
void __fastcall TInvoker::DoExecute()
{
	if(OnExecute != NULL) {
		OnExecute(this);
	}
}
//---------------------------------------------------------------------------
void __fastcall TInvoker::DoUndoing()
{
	if(OnUndoing != NULL) {
		OnUndoing(this);
	}
}
//---------------------------------------------------------------------------
void __fastcall TInvoker::DoUndo()
{
	if(OnUndo != NULL) {
		OnUndo(this);
	}
}
//---------------------------------------------------------------------------
void __fastcall TInvoker::DoRedoing()
{
	if(OnRedoing != NULL) {
		OnRedoing(this);
	}
}
//---------------------------------------------------------------------------
void __fastcall TInvoker::DoRedo()
{
	if(OnRedo != NULL) {
		OnRedo(this);
	}
}
//---------------------------------------------------------------------------
__fastcall TInvoker::TInvoker() :
	TObject(),
	FQueue(NULL),
	FHistory(NULL),
	FHistoryLimit(-1),
	FPosition(0),
	FOrigin(0),
	FOnExecuting(NULL),
	FOnExecute(NULL),
	FOnUndoing(NULL),
	FOnUndo(NULL),
	FOnRedoing(NULL),
	FOnRedo(NULL)
{
	Initialize();
}
//---------------------------------------------------------------------------
__fastcall TInvoker::~TInvoker()
{
	Uninitialize();
}
//---------------------------------------------------------------------------
void __fastcall TInvoker::Execute()
{
	if(!CanExecute()) {
		throw ECommandError("The program cannot execute any command.");
	}
	DoExecuting();
	try {
		Next->Execute();
	}
	catch(...) {
		FinishExecute();
		DoExecute();
		throw;
	}
	FinishExecute();
	DoExecute();
}
//---------------------------------------------------------------------------
bool __fastcall TInvoker::CanExecute()
{
	return Count > 0;
}
//---------------------------------------------------------------------------
void __fastcall TInvoker::Undo()
{
	if(!CanUndo()) {
		throw ECommandError("The program cannot undo any command.");
	}
	DoUndoing();
	try {
		NextUndo->Undo();
	}
	catch(...) {
		FinishUndo();
		DoUndo();
		throw;
	}
	FinishUndo();
	DoUndo();
}
//---------------------------------------------------------------------------
bool __fastcall TInvoker::CanUndo()
{
	return UndoCount > 0;
}
//---------------------------------------------------------------------------
void __fastcall TInvoker::Redo()
{
	if(!CanRedo()) {
		throw ECommandError("The program cannot redo any command.");
	}
	DoRedoing();
	try {
		NextRedo->Redo();
	}
	catch(...) {
		FinishRedo();
		DoRedo();
		throw;
	}
	FinishRedo();
	DoRedo();
}
//---------------------------------------------------------------------------
bool __fastcall TInvoker::CanRedo()
{
	return RedoCount > 0;
}
//---------------------------------------------------------------------------
void __fastcall TInvoker::Add(TCommand* ACommand)
{
	Queue->Add(ACommand);
}
//---------------------------------------------------------------------------
void __fastcall TInvoker::Clear()
{
	Queue->Clear();
}
//---------------------------------------------------------------------------
void __fastcall TInvoker::ClearHistory()
{
	RedoCount = 0;
	UndoCount = 0;
}
//---------------------------------------------------------------------------
void __fastcall TInvoker::ResetOrigin()
{
	FOrigin = FPosition;
}
//---------------------------------------------------------------------------
bool __fastcall TInvoker::IsModified()
{
	return Position != Origin;
}
//---------------------------------------------------------------------------
} // End namespace Moa_commandpattern
//---------------------------------------------------------------------------

