//---------------------------------------------------------------------------
#ifndef Moa_CommandPatternH
#define Moa_CommandPatternH
//---------------------------------------------------------------------------
#include <Classes.hpp>
//---------------------------------------------------------------------------
// Version: 2.0.0
// Date: 2012-07-12 UTC+09:00
// Author: Moochos
//---------------------------------------------------------------------------
namespace Moa_commandpattern {
//---------------------------------------------------------------------------
class ECommandError : public Exception
{
public:
	__fastcall ECommandError(const AnsiString& AMessage) : Exception(AMessage) {}
};
//---------------------------------------------------------------------------
enum TCommandResult
{
	crClearQueue,
	crClearUndo,
	crClearRedo,
	crCanUndo,
	crCanRedo,
};
//---------------------------------------------------------------------------
typedef Set<TCommandResult, crClearQueue, crCanRedo> TCommandResults;
//---------------------------------------------------------------------------
class TCommand : public TObject
{
private:
	TCommandResults FResults;
protected:
public:
	__fastcall TCommand();
	virtual __fastcall ~TCommand();
	virtual void __fastcall Execute() = 0;
	virtual void __fastcall Undo();
	virtual void __fastcall Redo();
	__property TCommandResults Results = {read=FResults, write=FResults};
};
//---------------------------------------------------------------------------
class TInvoker : public TObject
{
private:
	TObjectList* FQueue;
	TObjectList* FHistory;
	int FHistoryLimit;
	int FPosition;
	int FOrigin;
	TNotifyEvent FOnExecuting;
	TNotifyEvent FOnExecute;
	TNotifyEvent FOnUndoing;
	TNotifyEvent FOnUndo;
	TNotifyEvent FOnRedoing;
	TNotifyEvent FOnRedo;
	void __fastcall Initialize();
	void __fastcall Uninitialize();
	void __fastcall FinishExecute();
	void __fastcall FinishUndo();
	void __fastcall FinishRedo();
	TCommand* __fastcall GetNext();
	int __fastcall GetCount();
	TCommand* __fastcall GetNextUndo();
	int __fastcall GetUndoCount();
	void __fastcall SetUndoCount(int AUndoCount);
	TCommand* __fastcall GetNextRedo();
	int __fastcall GetRedoCount();
	void __fastcall SetRedoCount(int ARedoCount);
	int __fastcall GetHistorySize();
	void __fastcall SetHistoryLimit(int AHistoryLimit);
	__property TObjectList* Queue = {read=FQueue};
	__property TObjectList* History = {read=FHistory};
protected:
	virtual void __fastcall DoExecuting();
	virtual void __fastcall DoExecute();
	virtual void __fastcall DoUndoing();
	virtual void __fastcall DoUndo();
	virtual void __fastcall DoRedoing();
	virtual void __fastcall DoRedo();
public:
	__fastcall TInvoker();
	virtual __fastcall ~TInvoker();
	virtual void __fastcall Execute();
	virtual bool __fastcall CanExecute();
	virtual void __fastcall Undo();
	virtual bool __fastcall CanUndo();
	virtual void __fastcall Redo();
	virtual bool __fastcall CanRedo();
	virtual void __fastcall Add(TCommand* ACommand);
	virtual void __fastcall Clear();
	virtual void __fastcall ClearHistory();
	virtual void __fastcall ResetOrigin();
	virtual bool __fastcall IsModified();
	__property TCommand* Next = {read=GetNext};
	__property int Count = {read=GetCount};
	__property TCommand* NextUndo = {read=GetNextUndo};
	__property int UndoCount = {read=GetUndoCount, write=SetUndoCount};
	__property TCommand* NextRedo = {read=GetNextRedo};
	__property int RedoCount = {read=GetRedoCount, write=SetRedoCount};
	__property int HistorySize = {read=GetHistorySize};
	__property int HistoryLimit = {read=FHistoryLimit, write=SetHistoryLimit};
	__property int Position = {read=FPosition};
	__property int Origin = {read=FOrigin};
	__property TNotifyEvent OnExecuting = {read=FOnExecuting, write=FOnExecuting};
	__property TNotifyEvent OnExecute = {read=FOnExecute, write=FOnExecute};
	__property TNotifyEvent OnUndoing = {read=FOnUndoing, write=FOnUndoing};
	__property TNotifyEvent OnUndo = {read=FOnUndo, write=FOnUndo};
	__property TNotifyEvent OnRedoing = {read=FOnRedoing, write=FOnRedoing};
	__property TNotifyEvent OnRedo = {read=FOnRedo, write=FOnRedo};
};
//---------------------------------------------------------------------------
} // End namespace Moa_commandpattern
using namespace Moa_commandpattern;
//---------------------------------------------------------------------------
#endif

