//---------------------------------------------------------------------------
// Version: 2.2.0
// Date: 2012-07-12 UTC+09:00
// Author: Moochos
//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop
#include <ComCtrls.hpp>
#include <Math.hpp>
#include "Moa_AutoPointers.h"
#include "Moa_DynamicArrays.h"
#include "Moa_ListViews.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
//---------------------------------------------------------------------------
namespace Moa_listviews {
//---------------------------------------------------------------------------
void __fastcall TListViewItemUtilities::EnsureVisible(TListView* AListView, int AIndex, bool AAllowPartial)
{
	TListViewItemsUtilities::AssureValidIndex(AListView, AIndex);
	ListView_EnsureVisible(
		AListView->Handle,
		AIndex,
		AAllowPartial);
}
//---------------------------------------------------------------------------
UINT __fastcall TListViewItemUtilities::GetState(TListView* AListView, int AIndex)
{
	TListViewItemsUtilities::AssureValidIndex(AListView, AIndex);
	return ListView_GetItemState(
		AListView->Handle,
		AIndex,
		LVIS_CUT | LVIS_DROPHILITED | LVIS_FOCUSED | LVIS_SELECTED);
}
//---------------------------------------------------------------------------
void __fastcall TListViewItemUtilities::SetState(TListView* AListView, int AIndex, UINT AState)
{
	TListViewItemsUtilities::AssureValidIndex(AListView, AIndex);
	ListView_SetItemState(
		AListView->Handle,
		AIndex,
		AState,
		LVIS_CUT | LVIS_DROPHILITED | LVIS_FOCUSED | LVIS_SELECTED);
}
//---------------------------------------------------------------------------
bool __fastcall TListViewItemUtilities::IsCut(TListView* AListView, int AIndex)
{
	return (GetState(AListView, AIndex) & LVIS_CUT) != 0;
}
//---------------------------------------------------------------------------
void __fastcall TListViewItemUtilities::SetCut(TListView* AListView, int AIndex, bool ACut)
{
	UINT LItemState = GetState(AListView, AIndex);
	if(ACut) {
		LItemState |= LVIS_CUT;
	}
	else {
		LItemState &= ~LVIS_CUT;
	}
	SetState(AListView, AIndex, LItemState);
}
//---------------------------------------------------------------------------
bool __fastcall TListViewItemUtilities::IsDropHilighted(TListView* AListView, int AIndex)
{
	return (GetState(AListView, AIndex) & LVIS_DROPHILITED) != 0;
}
//---------------------------------------------------------------------------
void __fastcall TListViewItemUtilities::SetDropHilighted(TListView* AListView, int AIndex, bool ADropHilighted)
{
	UINT LItemState = GetState(AListView, AIndex);
	if(ADropHilighted) {
		LItemState |= LVIS_DROPHILITED;
	}
	else {
		LItemState &= ~LVIS_DROPHILITED;
	}
	SetState(AListView, AIndex, LItemState);
}
//---------------------------------------------------------------------------
bool __fastcall TListViewItemUtilities::IsFocused(TListView* AListView, int AIndex)
{
	return (GetState(AListView, AIndex) & LVIS_FOCUSED) != 0;
}
//---------------------------------------------------------------------------
void __fastcall TListViewItemUtilities::SetFocused(TListView* AListView, int AIndex, bool AFocused)
{
	UINT LItemState = GetState(AListView, AIndex);
	if(AFocused) {
		LItemState |= LVIS_FOCUSED;
	}
	else {
		LItemState &= ~LVIS_FOCUSED;
	}
	SetState(AListView, AIndex, LItemState);
}
//---------------------------------------------------------------------------
bool __fastcall TListViewItemUtilities::IsSelected(TListView* AListView, int AIndex)
{
	return (GetState(AListView, AIndex) & LVIS_SELECTED) != 0;
}
//---------------------------------------------------------------------------
void __fastcall TListViewItemUtilities::SetSelected(TListView* AListView, int AIndex, bool ASelected)
{
	UINT LItemState = GetState(AListView, AIndex);
	if(ASelected) {
		LItemState |= LVIS_SELECTED;
	}
	else {
		LItemState &= ~LVIS_SELECTED;
	}
	SetState(AListView, AIndex, LItemState);
}
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
TIntegerDynArray __fastcall TListViewItemsUtilities::AddItemsInRealMode(TListView* AListView, int ACount)
{
	TIntegerDynArray LIndexes;
	LIndexes.Length = ACount;
	for(int LI = 0; LI < ACount; ++LI) {
		LIndexes[LI] = AListView->Items->Add()->Index;
	}
	return LIndexes;
}
//---------------------------------------------------------------------------
TIntegerDynArray __fastcall TListViewItemsUtilities::AddItemsInVirtualMode(TListView* AListView, int ACount)
{
	AListView->Items->Count = AListView->Items->Count + ACount;
	return MakeIndexes(AListView->Items->Count - ACount, ACount);
}
//---------------------------------------------------------------------------
void __fastcall TListViewItemsUtilities::InsertItemsInRealMode(TListView* AListView, const TIntegerDynArray& AIndexes)
{
	if(AIndexes.Length == 0) {
		return;
	}
	for(int LI = 0; LI < AIndexes.Length; ++LI) {
		AListView->Items->Insert(AIndexes[LI]);
	}
}
//---------------------------------------------------------------------------
void __fastcall TListViewItemsUtilities::InsertItemsInVirtualMode(TListView* AListView, const TIntegerDynArray& AIndexes)
{
	if(AIndexes.Length == 0) {
		return;
	}
	AListView->Items->Count = AListView->Items->Count + AIndexes.Length;
	int LSourceIndex = AListView->Items->Count - AIndexes.Length;
	int LDestinationIndex = AListView->Items->Count;
	int LInsertCount = 0;
	while(--LSourceIndex < --LDestinationIndex) {
		if(LDestinationIndex == AIndexes[AIndexes.High - LInsertCount]) {
			++LInsertCount;
			++LSourceIndex;
		}
		else {
			Exchange(AListView, LSourceIndex, LDestinationIndex);
		}
	}
}
//---------------------------------------------------------------------------
void __fastcall TListViewItemsUtilities::DeleteItemsInRealMode(TListView* AListView, const TIntegerDynArray& AIndexes)
{
	if(AIndexes.Length == 0) {
		return;
	}
	for(int LI = 0; LI < AIndexes.Length; ++LI) {
		AListView->Items->Delete(AIndexes[AIndexes.High - LI]);
	}
}
//---------------------------------------------------------------------------
void __fastcall TListViewItemsUtilities::DeleteItemsInVirtualMode(TListView* AListView, const TIntegerDynArray& AIndexes)
{
	if(AIndexes.Length == 0) {
		return;
	}
	int LFocusIndex = -1;
	{
		int LIndex = AIndexes[0] - 1;
		int LItemCount = AListView->Items->Count;
		int LDeleteCount = 0;
		while(++LIndex < LItemCount) {
			UINT LItemState = TListViewItemUtilities::GetState(AListView, LIndex);
			if((LItemState & LVIS_FOCUSED) != 0) {
				LFocusIndex = EnsureRange(LIndex - LDeleteCount, -1, LItemCount - AIndexes.Length - 1);
			}
			if(LDeleteCount < AIndexes.Length && LIndex == AIndexes[LDeleteCount]) {
				++LDeleteCount;
			}
			else {
				TListViewItemUtilities::SetState(AListView, LIndex - LDeleteCount, LItemState);
			}
		}
		AListView->Items->Count = LItemCount - AIndexes.Length;
	}
	if(LFocusIndex >= 0) {
		TListViewItemUtilities::SetFocused(AListView, LFocusIndex, true);
	}
}
//---------------------------------------------------------------------------
void __fastcall TListViewItemsUtilities::ClearInRealMode(TListView* AListView)
{
	if(AListView->Items->Count == 0) {
		return;
	}
	AListView->Items->Clear();
}
//---------------------------------------------------------------------------
void __fastcall TListViewItemsUtilities::ClearInVirtualMode(TListView* AListView)
{
	if(AListView->Items->Count == 0) {
		return;
	}
	AListView->Items->Count = 0;
}
//---------------------------------------------------------------------------
void __fastcall TListViewItemsUtilities::ExchangeInRealMode(TListView* AListView, int AIndex1, int AIndex2)
{
	if(AIndex1 == AIndex2) {
		return;
	}
	TAutoPointer<TListItem> LItem = new TListItem(NULL);
	LItem->Assign(AListView->Items->Item[AIndex1]);
	AListView->Items->Item[AIndex1]->Assign(AListView->Items->Item[AIndex2]);
	AListView->Items->Item[AIndex2]->Assign(LItem);
}
//---------------------------------------------------------------------------
void __fastcall TListViewItemsUtilities::ExchangeInVirtualMode(TListView* AListView, int AIndex1, int AIndex2)
{
	AssureValidIndex(AListView, AIndex1);
	AssureValidIndex(AListView, AIndex2);
	if(AIndex1 == AIndex2) {
		return;
	}
	UINT LItemState = TListViewItemUtilities::GetState(AListView, AIndex1);
	TListViewItemUtilities::SetState(AListView, AIndex1, TListViewItemUtilities::GetState(AListView, AIndex2));
	TListViewItemUtilities::SetState(AListView, AIndex2, LItemState);
}
//---------------------------------------------------------------------------
void __fastcall TListViewItemsUtilities::MoveInRealMode(TListView* AListView, int ASourceIndex, int ADestinationIndex)
{
	if(ASourceIndex == ADestinationIndex) {
		return;
	}
	TAutoPointer<TListItem> LItem = new TListItem(NULL);
	LItem->Assign(AListView->Items->Item[ASourceIndex]);
	{
		int LStep = ASourceIndex < ADestinationIndex ? 1 : -1;
		int LIndex = ASourceIndex - LStep;
		while((LIndex += LStep) != ADestinationIndex) {
			AListView->Items->Item[LIndex]->Assign(AListView->Items->Item[LIndex + LStep]);
		}
	}
	AListView->Items->Item[ADestinationIndex]->Assign(LItem);
}
//---------------------------------------------------------------------------
void __fastcall TListViewItemsUtilities::MoveInVirtualMode(TListView* AListView, int ASourceIndex, int ADestinationIndex)
{
	AssureValidIndex(AListView, ASourceIndex);
	AssureValidIndex(AListView, ADestinationIndex);
	if(ASourceIndex == ADestinationIndex) {
		return;
	}
	UINT LItemState = TListViewItemUtilities::GetState(AListView, ASourceIndex);
	{
		int LStep = ASourceIndex < ADestinationIndex ? 1 : -1;
		int LIndex = ASourceIndex - LStep;
		while((LIndex += LStep) != ADestinationIndex) {
			TListViewItemUtilities::SetState(
				AListView,
				LIndex + LStep,
				TListViewItemUtilities::GetState(AListView, LIndex));
		}
	}
	TListViewItemUtilities::SetState(AListView, ADestinationIndex, LItemState);
}
//---------------------------------------------------------------------------
void __fastcall TListViewItemsUtilities::ReorderInRealMode(TListView* AListView, const TListReorderRecords& ARecords)
{
	if(ARecords.Length == 0) {
		return;
	}
	TAutoPointer<TObjectList> LList = new TObjectList();
	LList->Capacity = ARecords.Length;
	for(int LI = 0; LI < LList->Count; ++LI) {
		TAutoPointer<TListItem> LItem = new TListItem(NULL);
		LItem->Assign(AListView->Items->Item[ARecords[LI].SourceIndex]);
		LList->Add(LItem);
		LItem.Release();
	}
	for(int LI = 0; LI < LList->Count; ++LI) {
		AListView->Items->Item[ARecords[LI].DestinationIndex]->Assign(static_cast<TListItem*>(LList->Items[LI]));
	}
}
//---------------------------------------------------------------------------
void __fastcall TListViewItemsUtilities::ReorderInVirtualMode(TListView* AListView, const TListReorderRecords& ARecords)
{
	if(ARecords.Length == 0) {
		return;
	}
	DynamicArray<UINT> LItemStates;
	LItemStates.Length = ARecords.Length;
	for(int LI = 0; LI < LItemStates.Length; ++LI) {
		LItemStates[LI] = TListViewItemUtilities::GetState(
			AListView,
			ARecords[LI].SourceIndex);
	}
	for(int LI = 0; LI < LItemStates.Length; ++LI) {
		TListViewItemUtilities::SetState(
			AListView,
			ARecords[LI].DestinationIndex,
			LItemStates[LI]);
	}
}
//---------------------------------------------------------------------------
TIntegerDynArray __fastcall TListViewItemsUtilities::MakeIndexes(int AInitialIndex, int ACount)
{
	TIntegerDynArray LIndexes;
	LIndexes.Length = ACount;
	for(int LI = 0; LI < ACount; ++LI) {
		LIndexes[LI] = AInitialIndex + LI;
	}
	return LIndexes;
}
//---------------------------------------------------------------------------
int __fastcall TListViewItemsUtilities::Add(TListView* AListView)
{
	if(AListView->OwnerData) {
		return AddItemsInVirtualMode(AListView, 1)[0];
	}
	else {
		return AddItemsInRealMode(AListView, 1)[0];
	}
}
//---------------------------------------------------------------------------
TIntegerDynArray __fastcall TListViewItemsUtilities::AddItems(TListView* AListView, int ACount)
{
	if(AListView->OwnerData) {
		return AddItemsInVirtualMode(AListView, ACount);
	}
	else {
		return AddItemsInRealMode(AListView, ACount);
	}
}
//---------------------------------------------------------------------------
void __fastcall TListViewItemsUtilities::Insert(TListView* AListView, int AIndex)
{
	if(AListView->OwnerData) {
		InsertItemsInVirtualMode(AListView, MakeIndexes(AIndex, 1));
	}
	else {
		InsertItemsInRealMode(AListView, MakeIndexes(AIndex, 1));
	}
}
//---------------------------------------------------------------------------
void __fastcall TListViewItemsUtilities::InsertItems(TListView* AListView, const TIntegerDynArray& AIndexes)
{
	if(AListView->OwnerData) {
		InsertItemsInVirtualMode(AListView, AIndexes);
	}
	else {
		InsertItemsInRealMode(AListView, AIndexes);
	}
}
//---------------------------------------------------------------------------
TIntegerDynArray __fastcall TListViewItemsUtilities::InsertSucceedingItems(TListView* AListView, int AIndex, int ACount)
{
	TIntegerDynArray LIndexes = MakeIndexes(AIndex, ACount);
	if(AListView->OwnerData) {
		InsertItemsInVirtualMode(AListView, LIndexes);
	}
	else {
		InsertItemsInRealMode(AListView, LIndexes);
	}
	return LIndexes;
}
//---------------------------------------------------------------------------
void __fastcall TListViewItemsUtilities::Delete(TListView* AListView, int AIndex)
{
	DeleteItems(AListView, MakeIndexes(AIndex, 1));
}
//---------------------------------------------------------------------------
void __fastcall TListViewItemsUtilities::DeleteItems(TListView* AListView, const TIntegerDynArray& AIndexes)
{
	if(AListView->OwnerData) {
		DeleteItemsInVirtualMode(AListView, AIndexes);
	}
	else {
		DeleteItemsInRealMode(AListView, AIndexes);
	}
}
//---------------------------------------------------------------------------
void __fastcall TListViewItemsUtilities::DeleteSucceedingItems(TListView* AListView, int AIndex, int ACount)
{
	DeleteItems(AListView, MakeIndexes(AIndex, ACount));
}
//---------------------------------------------------------------------------
void __fastcall TListViewItemsUtilities::Clear(TListView* AListView)
{
	if(AListView->OwnerData) {
		ClearInVirtualMode(AListView);
	}
	else {
		ClearInRealMode(AListView);
	}
}
//---------------------------------------------------------------------------
void __fastcall TListViewItemsUtilities::Exchange(TListView* AListView, int AIndex1, int AIndex2)
{
	if(AListView->OwnerData) {
		ExchangeInVirtualMode(AListView, AIndex1, AIndex2);
	}
	else {
		ExchangeInRealMode(AListView, AIndex1, AIndex2);
	}
}
//---------------------------------------------------------------------------
void __fastcall TListViewItemsUtilities::ExchangeItems(TListView* AListView, const TListExchangeRecords& ARecords)
{
	for(int LI = 0; LI < ARecords.Length; ++LI) {
		Exchange(AListView, ARecords[LI].Index1, ARecords[LI].Index2);
	}
}
//---------------------------------------------------------------------------
void __fastcall TListViewItemsUtilities::Move(TListView* AListView, int ASourceIndex, int ADestinationIndex)
{
	if(AListView->OwnerData) {
		MoveInVirtualMode(AListView, ASourceIndex, ADestinationIndex);
	}
	else {
		MoveInRealMode(AListView, ASourceIndex, ADestinationIndex);
	}
}
//---------------------------------------------------------------------------
void __fastcall TListViewItemsUtilities::Reorder(TListView* AListView, const TListReorderRecords& ARecords)
{
	if(AListView->OwnerData) {
		ReorderInVirtualMode(AListView, ARecords);
	}
	else {
		ReorderInRealMode(AListView, ARecords);
	}
}
//---------------------------------------------------------------------------
void __fastcall TListViewItemsUtilities::SelectAll(TListView* AListView)
{
	AListView->SelectAll();
}
//---------------------------------------------------------------------------
void __fastcall TListViewItemsUtilities::ReverseSelection(TListView* AListView)
{
	for(int LI = 0; LI < AListView->Items->Count; ++LI) {
		TListViewItemUtilities::SetSelected(
			AListView,
			LI,
			!TListViewItemUtilities::IsSelected(AListView, LI));
	}
}
//---------------------------------------------------------------------------
void __fastcall TListViewItemsUtilities::ClearSelection(TListView* AListView)
{
	AListView->ClearSelection();
}
//---------------------------------------------------------------------------
int __fastcall TListViewItemsUtilities::GetSelectionIndex(TListView* AListView)
{
	return FindByState(AListView, LVIS_SELECTED);
}
//---------------------------------------------------------------------------
TIntegerDynArray __fastcall TListViewItemsUtilities::GetSelectionIndexes(TListView* AListView)
{
	TAutoPointer<TDynamicArrayBuilder<int> > LBuilder = new TDynamicArrayBuilder<int>();
	LBuilder->Capacity = AListView->SelCount;
	for(int LI = 0; LI < AListView->Items->Count; ++LI) {
		if(TListViewItemUtilities::IsSelected(AListView, LI)) {
			LBuilder->Add(LI);
		}
	}
	return LBuilder->Finish();
}
//---------------------------------------------------------------------------
void __fastcall TListViewItemsUtilities::SetSelectionIndexes(TListView* AListView, const TIntegerDynArray& AIndexes)
{
	if(AIndexes.Length == 0) {
		return;
	}
	int LSelectCount = 0;
	for(int LI = 0; LI < AListView->Items->Count; ++LI) {
		if(LI == AIndexes[LSelectCount]) {
			TListViewItemUtilities::SetSelected(AListView, LI, true);
			++LSelectCount;
		}
		else {
			TListViewItemUtilities::SetSelected(AListView, LI, false);
		}
	}
}
//---------------------------------------------------------------------------
int __fastcall TListViewItemsUtilities::GetSelectionCount(TListView* AListView)
{
	return FindByState(AListView, LVIS_SELECTED);
}
//---------------------------------------------------------------------------
int __fastcall TListViewItemsUtilities::GetFocusIndex(TListView* AListView)
{
	if(AListView->MultiSelect) {
		return AListView->ItemIndex;
	}
	else {
		return FindByState(AListView, LVIS_FOCUSED);
	}
}
//---------------------------------------------------------------------------
void __fastcall TListViewItemsUtilities::SetFocusIndex(TListView* AListView, int AIndex)
{
	return TListViewItemUtilities::SetFocused(AListView, AIndex, true);
}
//---------------------------------------------------------------------------
int __fastcall TListViewItemsUtilities::GetFocusCount(TListView* AListView)
{
	if(FindByState(AListView, LVIS_FOCUSED) >= 0) {
		return 1;
	}
	else {
		return 0;
	}
}
//---------------------------------------------------------------------------
int __fastcall TListViewItemsUtilities::FindByState(TListView* AListView, UINT AState)
{
	if(AState == LVIS_SELECTED && !AListView->MultiSelect) {
		return AListView->ItemIndex;
	}
	for(int LI = 0; LI < AListView->Items->Count; ++LI) {
		if((TListViewItemUtilities::GetState(AListView, LI) & AState) != 0) {
			return LI;
		}
	}
	return -1;
}
//---------------------------------------------------------------------------
TIntegerDynArray __fastcall TListViewItemsUtilities::FindItemsByState(TListView* AListView, UINT AState)
{
	TAutoPointer<TDynamicArrayBuilder<int> > LBuilder = new TDynamicArrayBuilder<int>();
	LBuilder->Capacity = CountByState(AListView, AState);
	for(int LI = 0; LI < AListView->Items->Count; ++LI) {
		if((TListViewItemUtilities::GetState(AListView, LI) & AState) != 0) {
			LBuilder->Add(LI);
		}
	}
	return LBuilder->Finish();
}
//---------------------------------------------------------------------------
int __fastcall TListViewItemsUtilities::CountByState(TListView* AListView, UINT AState)
{
	if(AState == LVIS_SELECTED) {
		return AListView->SelCount;
	}
	int LCount = 0;
	for(int LI = 0; LI < AListView->Items->Count; ++LI) {
		if((TListViewItemUtilities::GetState(AListView, LI) & AState) != 0) {
			++LCount;
		}
	}
	return LCount;
}
//---------------------------------------------------------------------------
void __fastcall TListViewItemsUtilities::InvalidateItems(TListView* AListView, const TIntegerDynArray& AIndexes)
{
	for(int LI = 0; LI < AIndexes.Length; ++LI) {
		AListView->UpdateItems(LI, LI);
	}
}
//---------------------------------------------------------------------------
void __fastcall TListViewItemsUtilities::InvalidateItems(TListView* AListView, int AStartIndex, int AEndIndex)
{
	AListView->UpdateItems(AStartIndex, AEndIndex);
}
//---------------------------------------------------------------------------
void __fastcall TListViewItemsUtilities::RepaintItems(TListView* AListView, int AStartIndex, int AEndIndex)
{
	InvalidateItems(AListView, AStartIndex, AEndIndex);
	AListView->Update();
}
//---------------------------------------------------------------------------
int __fastcall TListViewItemsUtilities::AssureValidIndex(TListView* AListView, int AIndex)
{
	if(AIndex < 0 || AIndex >= AListView->Items->Count) {
		throw EInvalidOperation("The index is invalid to the list view.");
	}
	return AIndex;
}
//---------------------------------------------------------------------------
} // End namespace Moa_listviews
//---------------------------------------------------------------------------

