//---------------------------------------------------------------------------
#ifndef Moa_ListsH
#define Moa_ListsH
//---------------------------------------------------------------------------
// Version: 2.3.0
// Date: 2012-07-12 UTC+09:00
// Author: Moochos
//---------------------------------------------------------------------------
#include "Moa_DynamicArrays.h"
//---------------------------------------------------------------------------
namespace Moa_lists {
//---------------------------------------------------------------------------
struct TListExchangeRecord
{
	int Index1;
	int Index2;
	__fastcall TListExchangeRecord();
	__fastcall TListExchangeRecord(int AIndex1, int AIndex2);
	__fastcall TListExchangeRecord(const TListExchangeRecord& ASource);
	TListExchangeRecord& __fastcall operator =(const TListExchangeRecord& ASource);
	bool __fastcall operator ==(const TListExchangeRecord& AOther) const;
	bool __fastcall operator !=(const TListExchangeRecord& AOther) const;
	TListExchangeRecord& __fastcall Reverse();
};
typedef DynamicArray<TListExchangeRecord> TListExchangeRecords;
//---------------------------------------------------------------------------
struct TListReorderRecord
{
	int SourceIndex;
	int DestinationIndex;
	__fastcall TListReorderRecord();
	__fastcall TListReorderRecord(int ASourceIndex, int ADestinationIndex);
	__fastcall TListReorderRecord(const TListReorderRecord& ASource);
	TListReorderRecord& __fastcall operator =(const TListReorderRecord& ASource);
	bool __fastcall operator ==(const TListReorderRecord& AOther) const;
	bool __fastcall operator !=(const TListReorderRecord& AOther) const;
	TListReorderRecord& __fastcall Reverse();
};
typedef DynamicArray<TListReorderRecord> TListReorderRecords;
//---------------------------------------------------------------------------
typedef int __fastcall (* TListAlgorithmsSortCompare)(int AIndex1, int AIndex2, void* AData);
//---------------------------------------------------------------------------
class TListAlgorithms
{
private:
	struct TSortRecord
	{
		int Index;
		TListAlgorithmsSortCompare Compare;
		void* Data;
	};
	struct TSlideRecord
	{
		int Index;
		bool Target;
	};
	static int __fastcall SortCompare(void* AItem1, void* AItem2);
	static void __fastcall Swap(int& AValue1, int& AValue2);
protected:
public:
	static TListReorderRecords __fastcall Sort(int ACount, TListAlgorithmsSortCompare ACompare, void* AData);
	static TListReorderRecords __fastcall ReverseOrder(int ACount);
	static TListReorderRecords __fastcall Slide(int ACount, const TIntegerDynArray& AIndexes, int AStep);
	static TListExchangeRecords __fastcall ReverseListExchangeRecords(const TListExchangeRecords& ARecords);
	static TListReorderRecords __fastcall ListExchangeRecordsToReorderRecords(const TListExchangeRecords& ARecords);
	static TListReorderRecords __fastcall ReverseReorderRecords(const TListReorderRecords& ARecords);
	static TListExchangeRecords __fastcall ReorderRecordsToListExchangeRecords(const TListReorderRecords& ARecords);
};
//---------------------------------------------------------------------------
class TListUtilities
{
private:
protected:
public:
	static void __fastcall InsertItems(TList* AList, const TIntegerDynArray& AIndexes);
	static void __fastcall InsertSucceedingItems(TList* AList, int AIndex, int ACount);
	static void __fastcall DeleteItems(TList* AList, const TIntegerDynArray& AIndexes);
	static void __fastcall DeleteSucceedingItems(TList* AList, int AIndex, int ACount);
	static void __fastcall ExchangeItems(TList* AList, const TListExchangeRecords& ARecords);
	static void __fastcall Reorder(TList* AList, const TListReorderRecords& ARecords);
	static void __fastcall ReverseOrder(TList* AList);
	static void* __fastcall Replace(TList* AList, int AIndex, void* AItem);
	static void __fastcall EnsureCapacity(TList* AList, int ACapacity);
};
//---------------------------------------------------------------------------
class TObjectListUtilities : public TListUtilities
{
private:
protected:
public:
	using TListUtilities::Reorder;
	static void __fastcall Reorder(TObjectList* AList, const TListReorderRecords& ARecords);
	using TListUtilities::Replace;
	static TObject* __fastcall Replace(TObjectList* AList, int AIndex, TObject* AItem);
};
//---------------------------------------------------------------------------
typedef void __fastcall (__closure * TListIndexEvent)(TObject* ASender, int AIndex);
typedef void __fastcall (__closure * TListIndexesEvent)(TObject* ASender, const TIntegerDynArray& AIndexes);
typedef void __fastcall (__closure * TListReorderEvent)(TObject* ASender, const TListReorderRecords& ARecords);
typedef void __fastcall (__closure * TListBlockEvent)(TObject* ASender, int AIndex, int ACount);
typedef void __fastcall (__closure * TListRangeEvent)(TObject* ASender, int AStartIndex, int AEndIndex);
//---------------------------------------------------------------------------
class TListInsertHistory : public TObject
{
private:
	TDynamicArrayBuilder<int>* FHistory;
	void __fastcall Initialize();
	void __fastcall Uninitialize();
	int __fastcall GetCapacity();
	void __fastcall SetCapacity(int ACapacity);
	__property TDynamicArrayBuilder<int>* History = {read=FHistory};
protected:
public:
	__fastcall TListInsertHistory();
	virtual __fastcall ~TListInsertHistory();
	virtual void __fastcall Add(int AIndex);
	virtual TIntegerDynArray __fastcall Finish();
	__property int Capacity = {read=GetCapacity, write=SetCapacity};
};
//---------------------------------------------------------------------------
class TListDeleteHistory : public TObject
{
private:
	TDynamicArrayBuilder<int>* FHistory;
	void __fastcall Initialize();
	void __fastcall Uninitialize();
	int __fastcall GetCapacity();
	void __fastcall SetCapacity(int ACapacity);
	__property TDynamicArrayBuilder<int>* History = {read=FHistory};
protected:
public:
	__fastcall TListDeleteHistory();
	virtual __fastcall ~TListDeleteHistory();
	virtual void __fastcall Add(int AIndex);
	virtual TIntegerDynArray __fastcall Finish();
	__property int Capacity = {read=GetCapacity, write=SetCapacity};
};
//---------------------------------------------------------------------------
class TListItemChangeHistory : public TObject
{
private:
	TDynamicArrayBuilder<int>* FHistory;
	static int __fastcall FinishSortCompare(void* AItem1, void* AItem2);
	void __fastcall Initialize();
	void __fastcall Uninitialize();
	int __fastcall GetCapacity();
	void __fastcall SetCapacity(int ACapacity);
	__property TDynamicArrayBuilder<int>* History = {read=FHistory};
protected:
public:
	__fastcall TListItemChangeHistory();
	virtual __fastcall ~TListItemChangeHistory();
	virtual void __fastcall Add(int AIndex);
	virtual TIntegerDynArray __fastcall Finish();
	__property int Capacity = {read=GetCapacity, write=SetCapacity};
};
//---------------------------------------------------------------------------
} // End namespace Moa_lists
using namespace Moa_lists;
//---------------------------------------------------------------------------
#endif

